#!/usr/bin/python

# Description

"""
Add My Tunes by Graeme Bell (http://graemebell.net) v2.0 Feb 2012

This script recursively adds a folder of music into iTunes so that if 
you sort your music by "Date Added", it will display each album 
separately, and the tracks should be listed in the correct order inside 
each album.

It does this by slowly moving one song at a time, and waiting for iTunes 
to keep up. Slightly slow, but it works. iTunes seems to benefit from a break 
every now and then so it can carry out some housekeeping tasks.

There is a no-GUI batch mode available. See amtconf.py for details."""

# System imports

import sys
import time
import os
import os.path
import threading
import Queue

import distutils.dir_util
import distutils.file_util

from Tkinter import *
import tkFileDialog
import tkMessageBox


# Program imports

import amtconf
import amtgui

from amthelper import debug, listdir_fullpath, sorted_nicely

# Main program


def main():
    if amtconf.gui_mode:    
        amtgui.gui_start()
    else:
        move_files_worker()



def directories_are_valid():

    # Sanity check directories

    if amtconf.from_dir == '' or amtconf.to_dir == '':
        amtgui.warning_box("Warning","You must choose both folders using the two top buttons.")
        return False

    if amtconf.from_dir == amtconf.to_dir:
        amtgui.warning_box("Warning","You chose the same folder twice as source & destination.\n\nPlease choose different folders.")
        return False

    return True


def move_files():

    if not directories_are_valid:
        return()

    #open cancel dialog, lock GUI, start file moving thread
    
    amtconf.q = Queue.Queue() # new interaction queue

    quitbox = amtgui.make_progress_box() # Allow user to cancel file transfer

    t = threading.Thread(target=move_files_worker) # Start file move thread
    amtconf.endThread = False
    t.start()

    amtgui.root.wait_window(quitbox) # GUI will lock until quitbox is closed by user/program

    amtconf.endThread = True # if we get here, the user hit quit or copying ended

    amtgui.done_box()

def recursively_explore(from_dir):

    folderStack= []   #Store files to move here. 

    #walk from bottom to top in reverse order by dir. 
    #pick out files in each dir and reverse sort. 
    for item in os.walk(from_dir, False):  #recursively descend
        name=item[0]  
        if os.path.isdir(name): # if it's a dir, sort & add the files, followed by the folder
            dir=name
            tmpDirContents = []
            for f in os.listdir(dir):
                if f == ".DS_Store": continue   # do not use 'break'! bypass stupid files on MacOS
                content = os.path.join(dir, f)  
                if os.path.isfile(content):
                    tmpDirContents.append(content)
            tmpDirContents=sorted_nicely(tmpDirContents)  # naturalsort
            debug(', '.join(tmpDirContents))
            if amtconf.sort_order == "NewestAtTop":   
                tmpDirContents.reverse() 
            folderStack.extend(tmpDirContents) #add sorted file list for this folder to the big list
            folderStack.append(name) # add folder to the big list

    return folderStack


def move_files_worker():

    # This function moves files into the auto-add folder slowly. 
    # Two stages. 1. build list of files 2. move them.

    #Get config variables 

    housekeeping_gap=amtconf.housekeeping_gap
    song_gap=amtconf.song_gap
    to_dir=amtconf.to_dir
    from_dir=amtconf.from_dir

    debug("housekeeping gap %d, song gap %d, to dir %s, from dir %s" % ( housekeeping_gap, song_gap, to_dir, from_dir ))

    folderStack = recursively_explore(from_dir)
        
    # now move/delete the list of files and folders
    filecount=0 # track number of transfers

    for src in folderStack:
        dst=os.path.join(to_dir,os.path.split(src)[1])  
        debug("Trying to move "+src+" to "+" "+to_dir)
        try:
            #If we've copied 'max_move' (e.g. 200) files/folders, pause.
            if filecount%amtconf.max_move == amtconf.max_move-1:  
                itunes_housekeeping_sleep(housekeeping_gap)
                                
            if amtconf.endThread: break 

            # Otherwise copy or move 1 file WITHOUT setting file permissions
            # This is important; there is a permissions/delete race with iTunes.

            if os.path.isfile(src) and (os.path.splitext(src)[1][1:] in amtconf.ok_types):
                amtconf.q.put(["UPDATE","Items done (%s): %d" % (amtconf.move_or_copy,filecount),"Transfer in progress."])
                distutils.file_util.copy_file(src,to_dir,0,0) #0,0 avoids race condition involving permissions operation & itunes

                if amtconf.move_or_copy=='Move':
                    os.remove(src) # nb. Don't replace this with a call to move_file; has no option to disable perms race cond

                filecount=filecount+1
                amtconf.q.put(["UPDATE","Items done (%s): %d" % (amtconf.move_or_copy,filecount),"Waiting for iTunes to add file"])
                #says moved/copied depending on mode. new in v2.  **tested*

                wait_for_itunes_to_copy(dst)

                time.sleep(song_gap) # add delay for each file to try to force iTunes to order things right

            else:
                debug("Skipping %s: unsuitable filetype or folder" % src)

        except IOError:
            debug("File/dir disappeared during copy - you may need to retry")

    amtconf.q.put(["DONE"])
    debug("Done.")
    
    from_dir = '' # reset the source folder, since it no longer exists

def wait_for_itunes_to_copy(dst):
    while os.path.exists(dst) and not amtconf.endThread :  
        debug("...waiting for itunes to copy/remove files")
        time.sleep(0.1)  # polling delay

def itunes_housekeeping_sleep(housekeeping_gap):
    sleepcount=housekeeping_gap
    while not amtconf.endThread and sleepcount > 0:
        amtconf.q.put(["UPDATE","Sleeping for %d" % sleepcount, "Sleeping to let iTunes catch up"])
        time.sleep(1)
        sleepcount = sleepcount -1

if __name__ == "__main__":
    main()


# License

""" Add My Tunes
    Copyright (C) 2011-2012 Graeme Bell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""



# Thanks

# Thanks to the following people for their interesting suggestions.
# Dan Shepard, Tycho van Dijk

# quick guide to program structure

# amt is for 'actions' - e.g. starting program, moving files. also handles thread control
# amtconf is for global state variables used by the file mover and by the GUI.
# amtgui is for the gui 
# amthelper is just a couple of simple helper functions.
# The main thread is a tkinter GUI thread. this can start a file moving thread, which runs parallel
# with the tkinter thread. the two threads communicate through a thread-safe queue structure.

# updates

# done
# v2.1
#url checking of version number  
#fixed the internationalisation bug permanently
#debug() in amthelper needed a UTF-8 encode to ensure that e.g. ascii-only outputs could cope with it. 
#this was secondary source of crashes in py2app version. debug should be set False when building dist.
#natural sort used to replace asciibetical sort. Tracks now ordered 1,2,14 rather than 1,14,2.
#made move_files_worker a bit neater.
#gave example of hardwired paths - should be sensible defaults for a macos user. 
#tidied up function names
#added nogui mode code and options
#gave example nogui setting for windows and macos


# v2 - faster polling. 0.1 instead of 0.5
# functions tidied up, placed in modules. 
# removed folder moving
# added filetype checking (only moves right files)
# moved sleep statement into 'ok' file move code block
# reverse order feature
# arguably, copy/move
# tested new sorting order, skipping over stuff, move/copy , messages in progress box, delay per file
# fixed positioning of all warning windows. tidied geometry and layout of main/processing/config windows. tested. 
# added radio buttons for sort order and move/copy, tested. 
# v55 internal.

